package FullSmoke::JobController::Roles::SingleModule;

use Moose::Role;
use Path::Tiny;
use FullSmoke::Util::CPANIndex qw/distinfo mod2dist/;
use Function::Parameters;
use FullSmoke::JobController; # get the exports too
use FullSmoke::Util::MooseHelpers;

with 'FullSmoke::JobController';

requires 'perl_name';

# One of these two is required
has_str_arg module => (
  required => 0,
  documentation => 'Module from CPAN to test'
);
has_str_arg dist_path => (
  required => 0,
  documentation => 'The path to the dist to install',
);

has_str_arg dist_file => (
  required => 0,
  documentation => 'Raw filename to record the job with.  Defaults to based off the dist_path argument',
  lazy => 1,
  builder => '_dist_file_from_path'
);
has_str_arg dist_author => (
  required => 0,
  default => scalar getpwuid($<),
  documentation => 'Author to record the test under, defaults to current user on machine'
);
has_str_arg dist_version => (
  required => 0,
  lazy => 1,
  builder => '_version_from_file',
  documentation => 'Version to record the test under, defaults to parsing from dist_path'
);
has_str_arg dist_name => (
  required => 0,
  lazy => 1,
  builder => '_dist_name_from_file',
  documentation => 'Name of the dist to record the test under, defaults to parsing from dist_path'
);

method _dist_file_from_path() {
  my $path = Path::Tiny->new($self->dist_path);
  return $path->basename;
}

method _dist_name_from_file() {
  my $filename = $self->dist_file;
  # Try to parse Foo-Bar-0.001_01.tar.gz
  if ($filename =~ /(.*)-[^-]+/) {
    return $1;
  } else {
    die "Non-standard filename, must provide --dist_name"
  }
}

method _version_from_file() {
  my $filename = $self->dist_file;
  # Try to parse Foo-Bar-0.001_01.tar.gz
  if ($filename =~ /^.*-([^-]+)(?:\.tar\.gz|\.tar\.bz2|\.tar\.xz|\.zip)$/i) {
    return $1;
  } else {
    die "Non-standard filename, must provide --dist_version"
  }
}

method BUILD(@args) {
  die "Must provide one of --module or --dist_path" unless $self->has_module || $self->has_dist_path;

  # Then we move to the next one
  $self->add_rolephase(cpan_module => (
    code => method() {
      if ($self->has_module) {
        return $self->make_job(
          CPANModule =>
            perl_name => $self->perl_name,
            distinfo(mod2dist($self->module)),
        );
      } else {
        return $self->make_job(
          CPANModule =>
            perl_name => $self->perl_name,
            dist_name => $self->dist_name,
            dist_file => $self->dist_file,
            version => $self->dist_version,
            author => $self->dist_author,
        );
      }
    }
  ));
}

1;