package FullSmoke::JobManager;

use Moose;
use MooseX::Storage;
use FullSmoke::Util::MooseHelpers;
use Function::Parameters qw/:std :modifiers/;
use Module::Runtime;

has_int_arg max_workers => (
  documentation => 'Max number of workers to have going at once',
);

has _phase => (
  isa => 'Int',
  is => 'rw',
  accessor => 'phase',
);

has _controller => (
  isa => 'FullSmoke::JobController',
  is => 'ro',
  traits => ['DoNotSerialize'],
  required => 1,
  reader => 'controller'
);

has _finished_jobs => (
  isa => 'HashRef[FullSmoke::Jobs]',
  is => 'ro',
  reader => 'finished_jobs',
  handles => {
    _had_job => 'exists',
    _done_job => 'set',
    _get_f_job => 'get'
  }
);

has _jobs => (
  isa => 'HashRef[FullSmoke::Jobs]',
  is => 'ro',
  reader => 'jobs',
  traits => ['Hash'],
  handles => {
    _have_job => 'exists',
    store_job => 'set',
    _get_job => 'get',
    _clear_job => 'delete',
    _jobs => 'values',
    active_jobs => 'count'
  }
);

has _workers => (
  isa => 'HashRef[HashRef]',
  is => 'ro',
  traits => ['Hash', 'DoNotSerialize'],
  lazy => 1,
  default => sub {[]},
  reader => 'workers',
  handles => {
    active_workers => 'count',
    _clear_worker => 'delete',
    _set_worker => 'set',
  }
);

{my $worker_id = 0; 
  method next_worker_id() {
    return $worker_id++;
  }
}

method done_job($job) {
  my $name = $job->canonical_name();
  $self->_done_job($name, $job);
  $self->_clear_job($name);
}

method has_job($name) {
  return $self->_have_job($name) || $self->_had_job($name);
}

method get_job($name) {
  return $self->_have_job($name) ? $self->_get_job($name) :
         $self->_had_job($name)  ? $self->_get_f_job($name) : undef;
}

method find_next_job() {
  my @jobs = $self->_jobs;

  for my $job (@jobs) {
    # Job is finished, move it to the other queue so we don't query it over and over
    if ($job->is_done) {
      $self->done_job($job->canonical_name, $job);
      next;
    }

    if ($job->is_built() || $job->is_moduleready()) {
      return $job;
    }
  }

  # Couldn't find an available job
  return;
}

method get_phase_jobs($phase_num) {
  my $phase = $self->controller->phase($phase_num);
  my @jobs = $phase->code->($self->controller);

  for my $job (@$phase) {
    my $name = $job->canonical_name();

    # Unless we already have the job, we'll add it to our queue
    unless ($self->has_job($name)) {
      $self->store_job($name, $job);
    }
  }
}

method make_job($name, %options) {
  my $class = use_module("FullSmoke::Jobs::$name");
  my $newjob = $class->new(%options);
  my $jobname = $newjob->canonical_name;
  
  if ($self->has_job($jobname)) {
    return $self->get_job($jobname);
  }

  # Not a dupe, return the new one
  return $newjob;
}

method tick() {
  # If we have too many active workers, don't add more
  return if ($self->active_workers >= $self->max_workers);

  while(my $job = $self->find_next_job) {
    my $id = $self->next_worker_id();
    
    my $future = $job->is_built() ? $job->setup() : $job->run();
    $self->_set_worker($id, {id => $id, job => $job, future => $future});

    $future->on_ready(fun ($future) {
      ### $self->log("Finished job...");
      $self->_delete_worker($id);
      $self->tick();
    });

    # Stop if we're full
    return if ($self->active_workers >= $self->max_workers);
  }

  # No next job!
  unless ($self->active_jobs > 0) {
    die "Phase finished!";
  }

  if ($self->active_workers == 0) {
    die "Workers are done, but we're looking for a job? wtf."
  }
}

1;