package FullSmoke::JobRole;

use v5.20;
use warnings;
use Moose::Role;
use Function::Parameters qw/:std :modifiers/;
use Future::AsyncAwait;
use Moose::Util::TypeConstraints;
use IO::Async::Function;
use Syntax::Keyword::Try;
use Time::HiRes qw/time/;
use MooseX::Storage;

with Storage;

has _timings => (
  is => 'ro',
  default => sub {[]},
  isa => 'ArrayRef',
  traits => ['Array'],
  handles => {
    '_add_timing' => 'push',
  },
  init_arg => undef,
);

method add_timing($name) {
  $self->_add_timing([$name, time()]);
}

has logs => (
  is => 'rw',
  isa => 'ArrayRef',
  traits => ['Array'],
  default => sub {[]},
  handles => {
    '_log' => 'push',
    'clear_logs' => 'clear',
  }
  init_arg => undef,
);

sub log {
  my ($self, @input) = @_;
  my $str = join('', @input);
  $str .= "\n" unless $str =~ /\n$/;
  $self->_log({time => time(), message => $str});
  print localtime()."", " $str";
};

# TODO part of the role?
has job_mgr => (
  is => 'ro',
  required => 1,
  weak_ref => 1,
  traits => ['DoNotSerialize']
);

requires 'setup';
around setup() {
  my $loop = $self->loop;
  die "Tried to setup after being setup" unless $self->_state() eq 'built';

  $self->_to_state("starting");
  my $function = IO::Async::Function->new(
    code => sub {
      my $ret;
      my $succ = 1;

      try {
        $ret = $self->$orig();
      } catch {
        $succ = 0;
        $self->log("FAILED: ", $@);
      }
      
      return {_ret => $ret, obj => $self->pack(), success => $succ};
    },
  );
  $loop->add($function);

  my $return = $function->call(args => []);

  $return->on_ready(fun($future) {
    my $data = $future->get();
    my $subp_obj = $self->unpack($data->{obj}); # makes a new obj

    $self->%{keys $subp_obj->%*} = $subp_obj->%*;

    if ($data->{success}) {
      # TODO use _ret to make dependencies!

      $self->_to_state('waiting');
    } else {
      $self->_to_state('failed');
    }

    $self->_dep_done(); # check if our deps are already done, and if so move our state
  });

  return $return;  
}

requires 'run';
has _future => (
  isa => 'Future',
  is => 'rw',
  predicate => 'has_future',
  traits => ['DoNotSerialize']
);

has 'loop' => (
  isa => 'IO::Async::Loop',
  is => 'ro',
  required => 1,
  traits => ['DoNotSerialize']
);

around run($orig, $self:) {
  my $loop = $self->loop;
  die "Tried to build future before ready" unless $self->is_ready();

  my $function = IO::Async::Function->new(
    code => sub {
      $self->clear_logs(); # sub process, so we want a clean set of logs
      my $ret;
      my $succ = 1;
      try {
        $ret = $self->$orig();
      } catch {
        $self->log("FAILED: ", $@);
        $succ = 0;
      }
      
      return {_ret => $ret, obj => $self->pack, success => $succ}
    }
  );
  $loop->add($function);

  my $return = $function->call(args => []);
  $self->_future($return); # TODO do I need to store this? maybe, because of result

  $self->_to_state('running');
  $return->on_ready(fun($future) {
    my $data = $future->get();
    my $subp_obj = $self->unpack($data->{obj}); # makes a new obj

    #$self->_log($data->{logs}->@*);
    $self->%{keys $subp_obj->%*} = $subp_obj->%*;

    if ($data->{success}) {
      $self->_to_state('done');
    } else {
      $self->_to_state('failed');
    }

    $self->_notify_waiters();
  });

  return $return;
}

my @states = qw[built starting waiting ready running done];
my %transitions = map {$states[$_] => $states[$_+1]} 0..$#states;
has _state => (
  isa => enum([@states, qw/failed depfailed/]),
  is => 'rw',
  default => "built",
  traits => ['DoNotSerialize'],
  init_arg => undef,
);

method is_built() {return $self->_state() eq 'built'}
method is_starting() {return $self->_state() eq 'starting'}
method is_waiting() {return $self->_state() eq 'waiting'}
method is_ready() {return $self->_state() eq 'ready'}
method is_done() {return $self->_state() eq 'done' || $self->_state eq 'depfailed' || $self->_state eq 'failed'}
method is_running() {return $self->_state() eq 'running'}
method is_failed() {return $self->_state() eq 'failed' || $self->_state eq 'depfailed'}

method _to_state($newstate) {
  if ($self->_state() eq $newstate) {
    warn "Double set to ".$self->_state();
  }
 
  if ($newstate ne $transitions{$self->_state} && $newstate ne 'failed' && $newstate ne 'depfailed') {
    die "Invalid state transition ".$self->_state()."=>$newstate"
  }

  $self->add_timing("changestate:$newstate");
  $self->_state($newstate);

  if ($newstate eq 'failed' || $newstate eq 'depfailed') {
    # tell our deps that we failed
    $self->_notify_waiters();
  }
}

#requires 'validate_options';
requires 'canonical_name';
has _dependencies => (
  traits => ['Array', 'DoNotSerialize'],
  is => 'ro',
  isa => 'ArrayRef[FullSmoke::Job]',
  default => sub {[]},
  handles => {
    _push_dep => 'push',
    _grep_deps => 'grep',
  }
);

has _working_deps => (
  traits => ['Array'],
  is => 'ro',
  isa => 'ArrayRef[HashRef]',
  default => sub {[]},
  handles => {
    _push_wdep => 'push'    
  }
);

method _check_deps() {
  return 0==0+scalar $self->_grep_deps(sub {!$_->is_done()})
}

method _failed_deps() {
  return 0+scalar $self->_grep_deps(sub {$_->is_failed()})
}

method register_dep(%options) {
  die "Can't add a dep while not starting" unless $self->_state eq 'starting';
  die "Must have job_type when registering a dependency" unless $options{job_type};

  my $job_type = $options{job_type};

  # Try to create it, if it works then we can save it
  # TODO also check if this can be compared for future searching/sorting
  my $testjob = "FullSmoke::Jobs::$job_type"->new(%options, loop => $loop);

  $self->_push_wdep(\%options);
}

method _reg_dep($dep) {
  die "Can't depend on anything but a job: ".ref($dep) unless $dep->isa("FullSmoke::Job");
  die "Can't add a dep while not starting" unless $self->_state eq 'starting';

  $self->_push_dep($dep);
  $dep->_on_done($self);
}

method _dep_done() {
  return unless $self->is_waiting();

  if ($self->_check_deps()) {
    if ($self->_failed_deps()) {
      $self->_to_state('depfailed');
    } else {
      $self->_to_state('ready');
    }
  }
}

has _waiters => (
  traits => ['Array'],
  isa => 'ArrayRef[FullSmoke::Job]',
  is => 'ro',
  default => sub {[]},
  handles => {
    _push_waiter => 'push',
  },
  init_arg => undef,
);

method _on_done($child) {
  weaken($child); # we only want a weak reference to our child jobs, they'll have a strong one to us
  $self->_push_waiter($child);
}

method _notify_waiters() {
  for my $wait ($self->_waiters->@*) {
    if (defined $wait) {
      $wait->_dep_done();
    } # otherwise we lost our weak reference, and don't care about it.
  }
}

# TODO really use this as async? not sure, might be better if it's not async
async sub result {
  my ($self) = @_;
  die "Can't get a result without being run" unless ($self->is_running || $self->is_done);
# TODO this needs to ensure the structure of the job result is what we want too
  if ($self->_future()->is_ready()) {
    return $self->_future->get();
  } else {
    return await $self->_future;
  }
}

# TODO in progress, needs some guards around these

method record_job() {
  # TODO serialize all information about the job, and any files
}

before ressurect_job($class: $data) {
  # Always guard some stuff here
  die "Can't ressurect on an object" if ref $class;
}

method ressurect_job($class: $data) {
  # TODO unserialize and init anything needed
}

1;
